<?php

namespace Slovocast\Infrastructure\Database;

use Slovocast\Infrastructure\Api\Database\ConnectionPoolInterface;
use Slovocast\Infrastructure\Api\Database\PooledConnectionInterface;
use SplObjectStorage;

class ConnectionPool implements ConnectionPoolInterface
{
    private int $waitTimeout;
    private SplObjectStorage $idleConnections;
    private SplObjectStorage $activeConnections;
    private ConnectionPoolConfig $config;

    public function __construct(ConnectionPoolConfig $config)
    {
        $this->idleConnections = new SplObjectStorage();
        $this->activeConnections = new SplObjectStorage();

        for ($i = 0; $i < $this->config->getTotalConnections(); $i++) {
            $pooledConnection = new PooledConnection($this->config->getDsnString());
            $this->idleConnections->attach($$pooledConnection);
        }

        $this->waitTimeout = $config->getPoolWaitTimeout();
    }

    public function getTotalIdleConnections(): int
    {
        return $this->idleConnections->count();
    }

    public function getTotalActiveConnections(): int
    {
        return $this->activeConnections->count();
    }

    public function hasIdleConnection(): bool
    {
        return $this->idleConnections->count() > 0;
    }

    public function setWaitTimeout(int $s): void
    {
        $this->waitTimeout = $s;
    }

    public function getWaitTimeout(): int
    {
        return $this->waitTimeout;
    }

    public function getConnectionLimit(): int
    {
        return $this->config->getTotalConnections();
    }

    /**
     * @TODO Throw an exception when a total timeout is exceeded. We do not want
     *       to get into an infinite loop.
     *
     * @return PooledConnectionInterface
     */
    public function getConnection(): PooledConnectionInterface
    {
        if (!$this->hasIdleConnection()) {
            \React\Async\delay((float) $this->getWaitTimeout());
            return this->getConnection();
        }

        $conn = $this->idleConnections->current();
        $this->idleConnections->detach($conn);
        $this->activeConnections->attach($conn);
        return $conn;
    }

    public function releaseConnection(PooledConnectionInterface $connection): void
    {
        if ($this->activeConnections->contains($connection)) {
            $this->activeConnections->detach($connection);
            $this->idleConnections->attach($connection);
        }
    }
}
